from abc import ABC, abstractmethod
import rstr
import random
import re

class ChromosomeFactory(ABC):
	"""
	Abstract Class to be inherited for implemention of different
	ways of generating initial population of chromosomes

	Instance variables :
	------------------
	noOfGenes : number of genes in each chromosome

	Methods :
	---------
	createChromosome() : Abstract method to be implemented by derived classes

	"""

	def __init__(self,noOfGenes):
		self.noOfGenes = noOfGenes

	@abstractmethod
	def createChromosome(self):
		"""
		Abstract method to be implemented by derived classes

		"""
		pass

class ChromosomeRegexFactory(ChromosomeFactory):
	"""
	Class derived from ChromosomeFactory, implements the method createChromosome()
	which generates initial population of candidates by using regex module in python
	on genes

	"""

	def __init__(self,noOfGenes,pattern,data_type=str):
		"""
		Parameters :
		------------

		noOfGenes : int ,  number of genes in each chromosome
		pattern : string containing individual genes
		data_type : datatype of each gene

		"""

		if noOfGenes <= 0:
			raise ValueError('No of genes cannot be negative or zero')
		if data_type not in [str,int,float]:
			raise ValueError('Invalid datatype given to Chromosome Regex Factory')
		try:
			re.compile(pattern)
		except re.error:
			raise ValueError('Invalid regex given to Chromosome Regex Factory')

		ChromosomeFactory.__init__(self,noOfGenes)
		self.data_type = data_type
		self.pattern = pattern
		

	def createChromosome(self):
		"""
		Generates a chromosome from given genes using python regex module

		Returns :
		---------
		chromosome : List containing individual genes of chromosome

		"""
		try:
			if self.data_type == int:
				chromosome = [int(rstr.xeger(self.pattern)) for i in range(self.noOfGenes)]
			elif self.data_type == float:
				chromosome = [float(rstr.xeger(self.pattern)) for i in range(self.noOfGenes)]
			else:
				chromosome = [rstr.xeger(self.pattern) for i in range(self.noOfGenes)]
			return chromosome
		except:
			raise Exception('Unable to convert all/some strings of given regex to type %s'%(type(self.data_type).__name__))
		

class ChromosomeRangeFactory(ChromosomeFactory):
	"""
	Class derived from ChromosomeFactory, implements the method createChromosome()
	which generates initial population of candidates by randomly sampling genes from a
	range of genes

	"""

	def __init__(self,noOfGenes,minValue,maxValue,duplicates=False):
		"""
		Parameters :
		-----------

		noOfGenes : int , number of genes in each chromosome
		minValue : int , lower bound of range
		maxValue : int , upper bound of range
		duplicates : boolean , indicates if gene can be repeated in chromosome
		data_type : datatype of each gene

		"""

		if noOfGenes <= 0 :
			raise ValueError('No of genes cannot be negative')

		if minValue > maxValue:
			raise ValueError('minValue cannot be greater than maxValue')

		if type(duplicates) != bool:
			raise ValueError('Invalid duplicated value given')
		
		ChromosomeFactory.__init__(self,noOfGenes)
		self.minValue = minValue
		self.maxValue = maxValue
		self.duplicates = duplicates
	
	def createChromosome(self):
		"""
		Generates a chromosome by randomly sampling genes from a given range

		Returns :
		---------
		chromosome : List of genes representing each chromosome

		"""
		try:
			if self.duplicates:
				chromosome = random.choices(range(self.minValue,self.maxValue+1), k = self.noOfGenes)
			else:
				chromosome = random.sample(range(self.minValue,self.maxValue+1), self.noOfGenes)
			return chromosome
		except:
			raise Exception('Unable to generated chromosome from given max %s min %s noOfGenes %s'%(self.minValue,self.maxValue,self.noOfGenes))




import random
import copy
import math
import operator

class SelectionHandlers:
	"""
	Class contains staticmethods for selection operation which are
	invoked by  Evolution.py module during selection operation
	
	Methods :
	--------
	random() : Randomly selects a fixed number of chromosomes from fitness mappings
	smallest() : Sorts and selects a fixed number of the smallest chromosomes from fitness mappings
	largest() : Sorts and selects a fixed number of the largest chromosomes from fitness mappings
	best() : Performs selection based on fitness type (min,max or equal).
				min -> smallest, max -> largest, equal -> least absolute difference
	tournament() : Performs tournament selection
	roulette() : Performs roulette based selection
	rank() : Ranks population based on fitness type, then performs roulette selection
	SUS() : Performs selection based on Stochastic Universal Sampling

	"""

	@staticmethod
	def random(fitness_mappings, ga):
		"""
		Randomly selects a fixed number of chromosomes from fitness mappings
	
		Parameters :
		----------
		fitness_mappings : list
					list storing population members and associated fitness values
		ga : Class reference
					Reference to GAEngine class

		Returns :
		----------
		Filtered population list based on selection method

		"""
		pop = ga.population.members
		return [random.choice(pop) for i in range(len(fitness_mappings)-math.ceil(ga.cross_prob * len(fitness_mappings)))]

	@staticmethod
	def smallest(fitness_mappings, ga):
		"""
		Sorts and selects a fixed number of the smallest chromosomes from fitness mappings
	
		Parameters :
		----------
		fitness_mappings : list
					list storing population members and associated fitness values
		ga : Class reference
					Reference to GAEngine class

		Returns :
		----------
		Filtered population list based on selection method
		
		"""
		new = sorted(fitness_mappings, key=operator.itemgetter(1))[:len(fitness_mappings)-math.ceil(ga.cross_prob * len(fitness_mappings))]
		return [i[0] for i in new]
		
	@staticmethod
	def largest(fitness_mappings, ga):
		"""
		Sorts and selects a fixed number of the largest chromosomes from fitness mappings
	
		Parameters :
		----------
		fitness_mappings : list
					list storing population members and associated fitness values
		ga : Class reference
					Reference to GAEngine class

		Returns :
		----------
		Filtered population list based on selection method
		
		"""
		new = sorted(fitness_mappings, key=operator.itemgetter(1), reverse=True)[:len(fitness_mappings)-math.ceil(ga.cross_prob * len(fitness_mappings))]
		return [i[0] for i in new]

	@staticmethod
	def best(fitness_mappings, ga):
		"""
		Performs selection based on fitness type (min,max or equal).
		min -> smallest, max -> largest, equal -> least absolute difference
	
		Parameters :
		----------
		fitness_mappings : list
					list storing population members and associated fitness values
		ga : Class reference
					Reference to GAEngine class

		Returns :
		----------
		Filtered population list based on selection method
		
		"""
		if ga.fitness_type == 'max':
			new = sorted(fitness_mappings, key=operator.itemgetter(1), reverse=True)[:len(fitness_mappings)-math.ceil(ga.cross_prob * len(fitness_mappings))]
		elif ga.fitness_type == 'min':
			new = sorted(fitness_mappings, key=operator.itemgetter(1))[:len(fitness_mappings)-math.ceil(ga.cross_prob * len(fitness_mappings))]
		elif ga.fitness_type[0] == 'equal':
			new = sorted(fitness_mappings, key=lambda x:abs(x[1]-ga.fitness_type[1]))[:len(fitness_mappings)-math.ceil(ga.cross_prob * len(fitness_mappings))]
		return [i[0] for i in new]
		
	@staticmethod
	def tournament(fitness_mappings, ga, tournsize):
		"""
		Performs tournament selection
	
		Parameters :
		----------
		fitness_mappings : list
					list storing population members and associated fitness values
		ga : Class reference
					Reference to GAEngine class

		Returns :
		----------
		Filtered population list based on selection method
		
		"""
		chosen = []
		for i in range(len(fitness_mappings)-math.ceil(ga.cross_prob * len(fitness_mappings))):
			aspirants = random.sample(fitness_mappings, tournsize)
			if ga.fitness_type == 'max':
				chosen.append(max(aspirants, key=operator.itemgetter(1)))
			elif ga.fitness_type == 'min':
				chosen.append(min(aspirants, key=operator.itemgetter(1)))
			elif ga.fitness_type[0] == 'equal':
				for x in range(tournsize):
					aspirants[x] = (aspirants[x][0],abs(aspirants[x][1] - ga.fitness_type[1]))
				chosen.append(min(aspirants, key=operator.itemgetter(1)))
		return [i[0] for i in chosen]
		
	@staticmethod
	def roulette(fitness_mappings, ga):
		"""
		Performs roulette based selection
	
		Parameters :
		----------
		fitness_mappings : list
					list storing population members and associated fitness values
		ga : Class reference
					Reference to GAEngine class

		Returns :
		----------
		Filtered population list based on selection method
		
		"""
		fitness = [i[1] for i in fitness_mappings]
		if ga.selection_handler == SelectionHandlers.roulette:
			if ga.fitness_type == 'max':
				pass
			elif ga.fitness_type == 'min':
				maxim = max(fitness)
				fitness = [maxim-x for x in fitness]
			elif ga.fitness_type[0] == 'equal':
				fitness = [abs(x-ga.fitness_type[1]) for x in fitness]
				maxim = max(fitness)
				fitness = [maxim-x for x in fitness]
		total_fit = float(sum(fitness))
		relative_fitness = [f/total_fit for f in fitness]
		probabilities = [sum(relative_fitness[:i+1]) for i in range(len(relative_fitness))]
		chosen = []
		pop = ga.population.members
		for n in range(len(fitness_mappings)-math.ceil(ga.cross_prob * len(fitness_mappings))):
			r = random.random()
			for (i, individual) in enumerate(pop):
				if r <= probabilities[i]:
					chosen.append(list(individual))
					break
		return chosen

	@staticmethod
	def rank(fitness_mappings, ga):
		"""
		Ranks population based on fitness type, then performs roulette selection
	
		Parameters :
		----------
		fitness_mappings : list
					list storing population members and associated fitness values
		ga : Class reference
					Reference to GAEngine class

		Returns :
		----------
		Filtered population list based on selection method
		
		"""
		if ga.fitness_type == 'max':
			new = [i[0] for i in sorted(fitness_mappings, key=operator.itemgetter(1))]
		elif ga.fitness_type == 'min':
			new = [i[0] for i in sorted(fitness_mappings, key=operator.itemgetter(1), reverse = True)]
		elif ga.fitness_type[0] == 'equal':
			new_dict = [(x[0],abs(x[1]-ga.fitness_type[1])) for x in fitness_mappings]
			new = [i[0] for i in sorted(new_dict, key=operator.itemgetter(1), reverse = True)]
		for i in range(len(new)):
			new[i]=(new[i],i+1)
		return SelectionHandlers.roulette(new, ga)

	@staticmethod
	def SUS(fitness_mappings, ga):
		"""
		Performs selection based on Stochastic Universal Sampling
	
		Parameters :
		----------
		fitness_mappings : list
					list storing population members and associated fitness values
		ga : Class reference
					Reference to GAEngine class

		Returns :
		----------
		Filtered population list based on selection method
		
		"""
		if ga.fitness_type == 'max':
			s_inds = sorted(fitness_mappings, key=operator.itemgetter(1), reverse=True)
		elif ga.fitness_type == 'min':
			s_inds = fitness_mappings
			maxim = (max(s_inds, key = operator.itemgetter(1)))[1]
			for x in range(len(s_inds)):
				s_inds[x] = (s_inds[x][0],maxim - s_inds[x][1])
			s_inds = sorted(fitness_mappings, key=operator.itemgetter(1), reverse=True)
		elif ga.fitness_type[0] == 'equal':
			s_inds = fitness_mappings
			for x in range(len(s_inds)):
				s_inds[x] = (s_inds[x][0],abs(s_inds[x][1] - ga.fitness_type[1]))
			maxim = (max(s_inds, key = operator.itemgetter(1)))[1]
			for x in range(len(s_inds)):
				s_inds[x] = (s_inds[x][0],maxim - s_inds[x][1])
			s_inds = sorted(fitness_mappings, key=operator.itemgetter(1), reverse=True)
		distance = sum([i[1] for i in fitness_mappings]) / float(ga.population.population_size)
		start = random.uniform(0, distance)
		points = [start + i*distance for i in range(len(fitness_mappings)-math.ceil(ga.cross_prob * len(fitness_mappings)))]
		chosen = []
		for p in points:
			i = 0
			sum_ = s_inds[i][1]
			while sum_ < p:
				i += 1
				sum_ += s_inds[i][1]
			chosen.append(s_inds[i][0])
		return chosen


class MutationHandlers:
	"""
	Class contains different  mutation handlers to be invoked by Evolution.py
	module during mutation operation 
	
	Methods :
	---------
	swap() : Performs mutation by swapping genes of chromosome in single point manner
	bitFlip() : Performs mutation by choosing one gene and flipping it to it's complement

	"""

	@staticmethod
	def swap(chromosome):
		"""
		Performs mutation operation on individual chromomsome by random selecting a
		point on chromosome and the exchanging genes from either side of that point 

		Parameters :
		-----------
		chromosome : List
					The chromosome on which mutation is to be performed

		Returns :
		---------
		Mutated Chromosome

		"""
		index = random.randint(0,len(chromosome)-2)
		newchrom = copy.copy(chromosome)
		newchrom[index], newchrom[index+1] = newchrom[index+1], newchrom[index]
		return newchrom

	# only when genes are binary
	@staticmethod
	def bitFlip(chromosome):
		"""
		Performs mutation by choosing one gene and flipping it to it's complement

		Parameters :
		-----------
		chromosome : List
					The chromosome on which mutation is to be performed

		Returns :
		---------
		Mutated Chromosome

		"""
		index = random.randint(0,len(chromosome)-1)
		chromosome[index] = type(chromosome[index])(not chromosome[index])
		return chromosome

class CrossoverHandlers:
	"""
	Class contains different static methods to be invoked by Evolution.py module  
	during crossover operation
	
	Method :
	-------

	distinct() : performs crossover between to chromosomes by appending genes from one chromosome 
				into other without duplicating the gene present in first chromosome 
	onePoint() : Performs one-point crossover on two chromosomes by swapping second halves of each chromosome
	twoPoint() : Performs two-point crossover on two chromosomes by swapping the mid part of each chromosome
	PMX() : Performs Partially Matched Crossover on two chromosomes
	OX() : Performs Ordered Crossover on two chromosomes

	"""

	@staticmethod
	def distinct(chromosome1,chromosome2):
		"""
		Performs crossover between to chromosomes by appending genes from one chromosome 
		into other without duplicating the gene present in first chromosome

		Parameters :
		------------
		chromosome1 : chromosome as one parent
		chromosome2 : chromosome as second parent

		Returns :
		----------
		Tuple containing two new offspring chromosomes produced by crossover 

		"""

		# Need to do some error handling here
		r = random.randint(1,len(chromosome1)-2)
		new_chromosome1 = chromosome1[:r]
		for i in chromosome2:
			if i not in new_chromosome1:
				new_chromosome1.append(i)
		new_chromosome2 = chromosome2[:r]
		for i in chromosome1:
			if i not in new_chromosome2:
				new_chromosome2.append(i)
		return new_chromosome1,new_chromosome2
		
	@staticmethod
	def onePoint(chromosome1, chromosome2):
		"""
		Performs one-point crossover on two chromosomes by swapping second halves of each chromosome

		Parameters :
		------------
		chromosome1 : chromosome as one parent
		chromosome2 : chromosome as second parent

		Returns :
		----------
		Tuple containing two new offspring chromosomes produced by crossover 

		"""
		size = min(len(chromosome1), len(chromosome2))
		r = random.randint(1, size - 1)
		new_chromosome1 = chromosome1[:r]+chromosome2[r:]
		new_chromosome2 = chromosome2[:r]+chromosome1[r:]
		return new_chromosome1, new_chromosome2

	@staticmethod
	def twoPoint(chromosome1, chromosome2):
		"""
		Performs two-point crossover on two chromosomes by swapping the mid part of each chromosome

		Parameters :
		------------
		chromosome1 : chromosome as one parent
		chromosome2 : chromosome as second parent

		Returns :
		----------
		Tuple containing two new offspring chromosomes produced by crossover 

		"""
		size = min(len(chromosome1), len(chromosome2))
		a = random.randint(1, size - 2)
		b = random.randint(1, size - 1)
		if b==a:
			b+=1
		elif b<a:
			a,b=b,a
		new_chromosome1 = chromosome1[:a]+chromosome2[a:b]+chromosome1[b:]
		new_chromosome2 = chromosome2[:a]+chromosome1[a:b]+chromosome2[b:]
		return new_chromosome1, new_chromosome2
		
	@staticmethod
	def PMX(chromosome1, chromosome2): # Partially Matched Crossover
		"""
		Performs Partially Matched Crossover on two chromosomes

		Parameters :
		------------
		chromosome1 : chromosome as one parent
		chromosome2 : chromosome as second parent

		Returns :
		----------
		Tuple containing two new offspring chromosomes produced by crossover 

		"""
		size = min(len(chromosome1), len(chromosome2))
		a = random.randint(1, size - 2)
		b = random.randint(1, size - 1)
		if b==a:
			b+=1
		elif b<a:
			a,b=b,a
		new_chromosome1 = chromosome1[:a]+chromosome2[a:b]+chromosome1[b:]
		new_chromosome2 = chromosome2[:a]+chromosome1[a:b]+chromosome2[b:]
		mapping1=chromosome2[a:b]
		mapping2=chromosome1[a:b]
		i = b
		while(i!=a):
			while new_chromosome1[i] in mapping1:
				new_chromosome1[i] = mapping2[mapping1.index(new_chromosome1[i])]
			while new_chromosome2[i] in mapping2:
				new_chromosome2[i] = mapping1[mapping2.index(new_chromosome2[i])]
			i = (i+1)%size
		return new_chromosome1, new_chromosome2
	
	@staticmethod
	def OX(chromosome1, chromosome2): # Ordered Crossover
		"""
		Performs Ordered Crossover on two chromosomes

		Parameters :
		------------
		chromosome1 : chromosome as one parent
		chromosome2 : chromosome as second parent

		Returns :
		----------
		Tuple containing two new offspring chromosomes produced by crossover 

		"""
		size = min(len(chromosome1), len(chromosome2))
		a = random.randint(1, size - 2)
		b = random.randint(1, size - 1)
		if b==a:
			b+=1
		elif b<a:
			a,b=b,a
		new_chromosome1 = [None]*(a)+chromosome1[a:b]+[None]*(size-b)
		new_chromosome2 = [None]*(a)+chromosome2[a:b]+[None]*(size-b)
		i,j,k = b,b,b
		while True:
			if chromosome2[i] not in new_chromosome1:
				new_chromosome1[j] = chromosome2[i]
				j = (j+1)%size
			if chromosome1[i] not in new_chromosome2:
				new_chromosome2[k] = chromosome1[i]
				k = (k+1)%size
			i = (i+1)%size
			if i==b or (j==a and k==a):
				break
		return new_chromosome1, new_chromosome2


class Fitness:
	"""
	Class contains staticmethods for calculating fitness score of an individual 

	Methods :
	--------
	addition() : calculates fitness scores of a candidate
	TSP() : Specific to Travelling Salesman Problem
			Calculates fitness score of a chromosome as sum of distances between each adjacent city in given chromosome

	"""

	@staticmethod

	def addition(chromosome):
		"""
		Calculates fitness score of a chromosome as sum of genes

		Parameters :
		-----------
		chromosome : list
					The chromosome whose fitness is to be calculated
		
		Returns :
		----------
		Fitness score 

		"""
		return sum(chromosome)
		
	@staticmethod
	def TSP(chromosome, matrix):
		"""
		Specific to Travelling Salesman Problem
		Calculates fitness score of a chromosome as sum of distances between each adjacent city in given chromosome

		Parameters :
		-----------
		chromosome : list
					Sequence of cities to be visited in order
		matrix : 2D list
					Gives distances between cities
		
		Returns :
		----------
		Fitness score 
		
		"""
		total = 0
		for i in range(len(chromosome)-1):
			total += matrix[chromosome[i]][chromosome[i+1]]
		return total



from pygenetic.ChromosomeFactory import *
from pygenetic import Utils, GAEngine
from keras.models import Sequential
from keras.layers import Dense
import numpy
import random
import copy

class ANNTopologyChromosomeFactory(ChromosomeFactory):
	"""
	ChromosomeFactory Class used to create ANN Chromosomes

	Instance variables :
	------------------
	neuronsPerLayer : list of possible number of neurons in the ANN to be used
	activations: list of possible number of activations to be used in ANN
	optimizers: list of possible number of optimizers to be used in ANN
	hiddenLayers: number of hidden layers to be used in the ANN(fixed)

	Methods :
	---------
	createChromosome() : method to be create ANN chromosome

	"""
	def __init__(self,hiddenLayers,neuronsPerLayer,activations,optimizers):
		self.neuronsPerLayer = neuronsPerLayer
		self.activations = activations
		self.optimizers = optimizers
		self.hiddenLayers = hiddenLayers

	def createChromosome(self):
		"""
		Generates ANN chromosome

		"""
		chromosome = []
		for i in range(self.hiddenLayers):
			chromosome.extend([random.choice(self.neuronsPerLayer), random.choice(self.activations)])
		chromosome.append(random.choice(self.activations))
		chromosome.append(random.choice(self.optimizers))
		return chromosome

class ANNTopologyEvolve:
	"""
	Class used to find best ANN topology using GA

	Instance variables :
	------------------
	X,Y: input and output values for ANN to train on
	population_size: GA population size
	loss: ANN loss to be used(eg:'binary_crossentropy')
	metrics: ANN metrics to be used (eg:'accuracy')
	epochs: Number of ANN epochs
	batch_size: ANN Batch size  
	neuronsPerLayer : list of possible number of neurons in the ANN to be used
	activations: list of possible number of activations to be used in ANN
	optimizers: list of possible number of optimizers to be used in ANN
	hiddenLayers: number of hidden layers to be used in the ANN(fixed)

	Methods :
	---------
	evolve(noOfGenerations) : used to find best ANN topology using GA evolutions

	"""
	def __init__(self,X,Y,hiddenLayers,population_size=10,neuronsPerLayer=[2,5,10,12],activations=['relu','sigmoid'],optimizers=['adam'],loss='binary_crossentropy',metrics='accuracy',epochs=30,batch_size=10):
		self.X = X
		self.Y = Y
		if hiddenLayers < 1:
			raise Exception('No of hiddenLayers should be greater than zero')
		self.hiddenLayers = hiddenLayers
		self.input_dim = len(X[0])
		self.neuronsPerLayer = neuronsPerLayer
		self.activations = activations
		self.optimizers = optimizers
		self.loss = loss
		self.metrics = metrics
		self.epochs = epochs
		self.batch_size = batch_size
		self.population_size = population_size
		self.factory = ANNTopologyChromosomeFactory(hiddenLayers,neuronsPerLayer,activations,optimizers)

	def evolve(self,noOfGenerations=100):
		ga = GAEngine.GAEngine(self.factory,population_size=self.population_size,fitness_type='min')
		ga.setFitnessHandler(ANNTopologyEvolve.fitness,self.X,self.Y,self.hiddenLayers,self.input_dim,self.loss,self.metrics,self.epochs,self.batch_size)
		ga.setSelectionHandler(Utils.SelectionHandlers.best)
		ga.addCrossoverHandler(Utils.CrossoverHandlers.onePoint,1)
		ga.addMutationHandler(ANNTopologyEvolve.mutation,1,self.neuronsPerLayer,self.activations,self.optimizers)
		ga.evolve(noOfGenerations)

	@staticmethod
	def fitness(chromosome,X,Y,hidden_layers,input_dim,loss,metrics,epochs,batch_size):
		# create model
		model = Sequential()
		model.add(Dense(chromosome[0], input_dim= input_dim, activation=chromosome[1]))
		for i in range(hidden_layers-1):
			model.add(Dense(chromosome[i+2], activation=chromosome[i+3]))
		model.add(Dense(1, activation=chromosome[len(chromosome)-2]))
		# Compile model
		model.compile(loss=loss, optimizer=chromosome[len(chromosome)-1], metrics=[metrics])
		# Fit the model
		model.fit(X, Y, epochs=epochs, batch_size=batch_size,verbose=1)
		# evaluate the model
		scores = model.evaluate(X, Y)
		print(scores)
		print("\n%s: %.2f%%" % (model.metrics_names[1], scores[1]*100))
		print('Loss', scores[0])
		return scores[0]

	@staticmethod
	def mutation(chromosome,neuronsPerLayer,activations,optimizers):
		r = random.randint(0,len(chromosome)-1)
		newchrom = copy.copy(chromosome)
		if r == len(chromosome) - 1:
			newchrom[r] = random.choice(optimizers)
		elif r%2 == 1 or r == len(chromosome)-2:
			newchrom[r] = random.choice(activations)
		elif r%2 == 0:
			newchrom[r] = random.choice(neuronsPerLayer)
		return newchrom



import random
from pygenetic import Utils
import math
import numpy as np

class SimpleGA:
	"""
	This Class is the driver program which contains and invokes the operators used in Genetic algorithm
	This class can be invoked to implement a non-generic genetic solution 
	SimpleGA keeps track of specific type of operators the user has specified for running the algorithm

	Methods :
	----------
	create_initial_population() : Generates initial members of population by randomly generating chromosomes based on range
	doCrossover() : Performs crossover by calling specific utility function based on crossover_handler
	doMutation() : Performs mutation by calling specific utility function based on mutation_handler
	calculateFitness() : Returns fitness associated with chromosome passed as argument
	generateFitnessMappings() : Generates a list of population members and associated fitnesses
	handleSelection() : Generates fitness mappings and performs selection by calling specific utility function based on selection_handler
	evolve() : Performs evolution for specified number of iterations

	"""
	def __init__(self,minValue,maxValue,noOfGenes,fitness_func,duplicates=False,population_size=100,cross_prob=0.8,mut_prob=0.1,crossover_handler='onePoint',mutation_handler='swap',selection_handler='best',fitness_type='max'):
		self.minValue = minValue
		self.maxValue = maxValue
		self.noOfGenes = noOfGenes
		self.duplicates = duplicates
		self.population_size = population_size
		self.create_initial_population()
		self.cross_prob = cross_prob
		self.mut_prob = mut_prob
		self.crossover_handler = crossover_handler
		self.mutation_handler = mutation_handler
		self.selection_handler = selection_handler
		self.fitness_type = fitness_type
		if type(self.fitness_type) == str:
			if self.fitness_type == 'max':
				self.best_fitness = None, float("-inf")
			elif self.fitness_type == 'min':
				self.best_fitness = None, float("inf")
		elif type(self.fitness_type) == tuple or type(self.fitness_type) == list:
			if self.fitness_type[0] == 'equal':
				self.best_fitness = None, float("inf")
		self.fitness_func = fitness_func

	def create_initial_population(self):
		"""
		Generates initial members of population by randomly generating chromosomes based on range

		"""
		self.population = []
		for i in range(self.population_size):
			if self.duplicates:
				chromosome = random.choices(range(self.minValue,self.maxValue+1), k = self.noOfGenes)
			else:
				chromosome = random.sample(range(self.minValue,self.maxValue+1), self.noOfGenes)
			self.population.append(chromosome)
		print("Initial population: ", self.population)

	def doCrossover(self, chromosome1, chromosome2):
		"""
		Performs crossover by calling specific utility function based on crossover_handler

		Parameters :
		----------
		chromosome1 : chromosome as one parent
		chromosome2 : chromosome as second parent

		Returns :
		----------
		Tuple containing two new offspring chromosomes produced by crossover 

		"""
		if self.crossover_handler == 'distinct':
			return Utils.CrossoverHandlers.distinct(chromosome1,chromosome2)
		elif self.crossover_handler == 'onePoint':
			return Utils.CrossoverHandlers.onePoint(chromosome1,chromosome2)
		elif self.crossover_handler == 'twoPoint':
			return Utils.CrossoverHandlers.twoPoint(chromosome1,chromosome2)
		elif self.crossover_handler == 'PMX':
			return Utils.CrossoverHandlers.PMX(chromosome1,chromosome2)
		elif self.crossover_handler == 'OX':
			return Utils.CrossoverHandlers.OX(chromosome1,chromosome2)
		else:
			raise Exception('Invalid Crossover given as input')

	def doMutation(self, chromosome1):
		"""
		Performs mutation by calling specific utility function based on mutation_handler

		Parameters :
		----------
		chromosome1 : chromosome as subject

		Returns :
		----------
		Mutated chromosome

		"""
		if self.mutation_handler == 'swap':
			return Utils.MutationHandlers.swap(chromosome1)
		elif self.mutation_handler == 'bitFlip':
			return Utils.MutationHandlers.bitFlip(chromosome1)
		else:
			raise Exception('Invalid Mutation given as input')

	def calculateFitness(self,chromosome):
		"""
		Returns fitness associated with chromosome passed as argument

		Parameters :
		----------
		chromosome : chromosome whose fitness is to be calculated

		Returns :
		----------
		Fitness of that chromosome

		"""
		return self.fitness_func(chromosome)

	def generateFitnessMappings(self):
		"""
		Generates a list of population members and associated fitnesses

		"""
		self.fitness_mappings = [(member, self.calculateFitness(member)) for member in self.population]
		print(self.fitness_mappings)	

	def handle_selection(self):
		"""
		Generates fitness mappings and performs selection by calling specific utility function based on selection_handler

		Returns :
		----------
		Filtered population list based on selection method

		"""
		self.generateFitnessMappings()
		if self.fitness_type == 'max':
			self.fitness_mappings.sort(key=lambda x:x[1],reverse=True)
			self.best_fitness = self.fitness_mappings[0]
		elif self.fitness_type == 'min':
			self.fitness_mappings.sort(key=lambda x:x[1])
			self.best_fitness = self.fitness_mappings[0]
		elif self.fitness_type[0] == 'equal':
			self.fitness_mappings.sort(key=lambda x:abs(x[1]-self.fitness_type[1]))
			self.best_fitness = self.fitness_mappings[0]
		else:
			raise Exception('Invalid fitness type')

		if self.selection_handler == 'best':
			return Utils.SelectionHandlers.best(self.fitness_mappings,self)
		elif self.selection_handler == 'roulette':
			return Utils.SelectionHandlers.roulette(self.fitness_mappings,self)
		elif self.selection_handler == 'rank':
			return Utils.SelectionHandlers.rank(self.fitness_mappings,self)
		else:
			raise Exception('Invalid Mutation given as input')

	def evolve(self,noOfIterations=50):
		"""
		Performs the evolution as many times as number of iterations specified by user or terminates
		if optimal solution is found.

		Parameters :
		-----------
		noOfIterations : int
						default value : 50

		Outline of Algorithm :
		---------------------
		Selection : fittest members of population are selected by invoking 
					selection handler in Utils.py module
		Crossover : A probability score is generated from fitness value of each chromosome
					Chromosomes for crossover are selected based on this probablity score 
					of each chromosome
					Crossover is performed by invoking a crossover handler from Utils.py
		Mutation : The genes to be mutated are chosen randomly
					Once indexes of Chromomsome / genes to be mutated are determined, a mutation
					handler from Utils.py module is invoked 

		Each iteration consists of repeating the above operations until an optimal solution 
		determined by fitness threshold is reached  or number of iterations specified are complete

		"""
		for i in range(noOfIterations):
			new_population = self.handle_selection()
			print("*** Members left after selection = ",len(new_population))
			print("Best member = ",self.best_fitness[0])
			print("Best fitness = ",self.best_fitness[1])
			if self.fitness_type[0] == 'equal':
				if self.best_fitness[1] == self.fitness_type[1]:
					print('Solved')
					return 1
			fitnesses = []
			total = 0
			for chromosome in self.fitness_mappings:
				fitness = chromosome[1]
				if fitness == 0:
					fitness = random.uniform(0.01, 0.02)
				total += fitness
				fitnesses.append(fitness)
			print(fitnesses)	
			p = [ elem/total for elem in fitnesses]
			n = math.ceil(self.cross_prob * len(p))
			if n %2 == 1:
				n -= 1
				self.population.append(self.population[0])

			crossover_indexes = np.random.choice(len(p),n,p=p, replace=False)

			crossover_chromosomes = [ self.population[index] for index in crossover_indexes]

			for i in range(0,len(crossover_chromosomes)-1,2):
				father,mother = crossover_chromosomes[i], crossover_chromosomes[i+1]
				child1, child2 = self.doCrossover(father,mother)
				new_population.extend([child1,child2])
			mutation_indexes = np.random.choice(len(new_population),int(self.mut_prob*len(p)), replace=False)
			for index in mutation_indexes:
				new_population[index] = self.doMutation(new_population[index])
			self.population = new_population
			new_population = []
		print("Best fitness in this generation = ", self.best_fitness)
		print("Top 10 fitnesses in this generation = ", self.fitness_mappings[:10])



import matplotlib
import os
if 'TRAVIS' in os.environ:
    print('Warning: no DISPLAY environment variable found. Using matplotlib non-interactive Agg backend')
    matplotlib.use('Agg')
import matplotlib.pyplot as plt

class Statistics:
	"""
	Class to generate Statistics on operation of Genetic Algorithm

	Instance Members:
	-----------------
	statistic_dict : A dictionary storing different statistics mapped to list storing each generation data
				Stats stored are best-fitness, worst-fitness, avg-fitness, diversity and mutation-rate

	"""


	def __init__(self):
		self.statistic_dict = {'best-fitness':[],'worst-fitness':[],'avg-fitness':[],'diversity':[],'mutation_rate':[]}

	def add_statistic(self,statistic,value):
		"""
		Appends a value to specified statistic, usually called after each iteration

		Parameters :
		------------
		statistic : The statistic for which the value is relevant and is to be appended
		value : Th evalue to be appended

		"""
		if statistic in self.statistic_dict:
			self.statistic_dict[statistic].append(value)
		else:
			self.statistic_dict[statistic] = [value]


	def plot(self):
		"""
		Generates a line graph for each statistic to display change over iterations

		"""
		fig,ax = plt.subplots()
		ax.set_xlabel('Generation')
		ax.set_ylabel('Statistic')
		for statistic in self.statistic_dict:
			ax.plot(range(1,len(self.statistic_dict[statistic])+1),self.statistic_dict[statistic],label=statistic)
		fig.legend(loc='upper left')
		return fig

	def plot_statistics(self,statistics):
		"""
		Generates a line graph for list of specified statistics to display change over iterations

		Parameters :
		----------
		statistics : A list of statistic names whose variation is to be shown

		"""
		fig,ax = plt.subplots()
		ax.set_xlabel('Generation')
		ax.set_ylabel('Statistic')
		for statistic in statistics:
			ax.plot(range(1,len(self.statistic_dict[statistic])+1),self.statistic_dict[statistic],label=statistic)
		fig.legend(loc='upper left')
		return fig

	def plot_statistic(self,statistic):
		"""
		Generates a line graph for specified statistic to display change over iterations

		Parameters :
		----------
		statistic : The statistic name whose variation is to be shown

		"""
		fig,ax = plt.subplots()
		ax.set_xlabel('Generation')
		ax.set_ylabel('Statistic')
		ax.plot(range(1,len(self.statistic_dict[statistic])+1),self.statistic_dict[statistic],label=statistic)
		fig.legend(loc='upper left')
		return fig



from abc import ABC, abstractmethod
import random
import math
from pygenetic import Utils
import numpy as np

class BaseEvolution(ABC):
	"""
	Abstract class to be inherited to implement the specific evolution procedure

	Methods :
	---------
	evolve() : abstract method to be implemeted by derived classes

	"""

	def __init__(self):
		pass

	@abstractmethod
	def evolve(self,ga):
		"""
		Abstract method to be implemeted by derived classes
		
		Parameters :
		-------------
		ga : reference to the GAEngine object

		"""

		pass

class StandardEvolution(BaseEvolution):
	"""
	Class inherits from BaseEvolution and contains implementations of abstract 
	evolution method in BaseEvolution

	Instance Members :
	------------------

	pyspark : boolean to indicated if parallelization should be supported by using pyspark

	"""
	def __init__(self,pyspark=False):
		self.pyspark = pyspark

	def __evolve_normal(self,ga):

		"""
		Private method which performs an iteration

		Parameters :
		-------------
		ga : reference to the GAEngine object

		Outline of Algorithm :
		---------------------
		Selection : fittest members of population are selected by invoking 
					selection handler in Utils.py module
		Crossover : A probability score is generated from fitness value of each chromosome
					Chromosomes for crossover are selected based on this probablity score 
					of each chromosome
					Crossover is performed by invoking a crossover handler from Utils.py
		Mutation : If adaptive mutation is set then average square deviation of fitness values
					is used for determining (the indexes of chromosome)/genes to be mutated 
					If false then genes to be mutated are chosen randomly
					Once indexes of Chromomsome / genes to be mutated are determined, a mutation
					handler from Utils.py module is invoked 

		Each iteration consists of repeating the above operations until an optimal solution 
		determined by fitness threshold is reached  or number of iterations specified are complete

		"""

		# get (1-r) * cross_prob new members
		ga.population.new_members = ga.handle_selection()

		if ga.fitness_type[0] == 'equal':
			if ga.best_fitness[1] == ga.fitness_type[1]:
				return 1

		fitnesses = []
		total_fitness = 0
		for chromosome in ga.fitness_mappings:
			fitness = chromosome[1]
			if fitness <= 0:
				fitness = random.uniform(0.01, 0.02)
			total_fitness += fitness
			fitnesses.append(fitness)

		p = [ elem/total_fitness for elem in fitnesses]

		n = math.ceil(ga.cross_prob * len(p))
		if n %2 == 1:
			n -= 1
			ga.population.new_members.append(ga.population.members[random.randint(0,len(p)-1)])

		crossover_indexes = np.random.choice(len(p),n,p=p, replace=False)

		crossover_chromosomes = [ ga.population.members[index] for index in crossover_indexes]

		for i in range(0,len(crossover_chromosomes)-1,2):
			father,mother = crossover_chromosomes[i], crossover_chromosomes[i+1]
			crossoverHandler = ga.chooseCrossoverHandler()
			child1, child2 = ga.doCrossover(crossoverHandler,father,mother)
			ga.population.new_members.extend([child1,child2])


		mutation_indexes = np.random.choice(len(ga.population.new_members),int(ga.mut_prob*len(p)), replace=False)

		for index in mutation_indexes:
			mutationHandler = ga.chooseMutationHandler()
			ga.population.new_members[index] = ga.doMutation(mutationHandler,ga.population.new_members[index])

		return 0

	def __evolve_pyspark(self,ga):
		"""
		Private method which performs iterations using pyspark

		Parameters :
		-------------
		ga : reference to the GAEngine object

		Outline of Algorithm:
		---------------------
		Selection :
			- Initial population rdd creation
			- Each chromosome is mapped to key-value pair (chromosome , fitness)
			- Depending on type of fitness type fitest chromosomes are selected for crossover
			- The mapped chromosome rdd is transformed to another rdd containing probabilities for crossover 
				which are calculated by normalizing fitness values of each chromosome  

		Crossover :
			- Rdd containing chromosome pairs to be crossed over are created using the probability score calculated
			- New chromosomes are appended to population
		Mutation:
			- Mutation chromosome rdds are created which are transformed to rdd (index, mutated chromosome)
			- Then collect action is applied to get the mutated chromsomes and used  to replace the corresponding members of population

		"""
		from pyspark import SparkContext
		sc = SparkContext.getOrCreate()
		chromosomes_rdd = sc.parallelize(ga.population.members)
		# Fitness Value Mapping and making selection
		mapped_chromosomes_rdd = chromosomes_rdd.map(lambda x: (x,ga.calculateFitness(x)))
		if ga.selection_handler == Utils.SelectionHandlers.best:
			if type(ga.fitness_type) == str:
				if ga.fitness_type == 'max':
					selected_chromosomes = mapped_chromosomes_rdd.top(len(ga.population.members)-math.ceil(ga.cross_prob * len(ga.population.members)),key=lambda x: x[1])
					ga.best_fitness = selected_chromosomes[0]
					selected_chromosomes = [ x[0] for x in selected_chromosomes]
				elif ga.fitness_type == 'min':
					selected_chromosomes = mapped_chromosomes_rdd.takeOrdered(len(ga.population.members)-math.ceil(ga.cross_prob * len(ga.population.members)),key=lambda x: x[1])
					ga.best_fitness = selected_chromosomes[0]
					selected_chromosomes = [ x[0] for x in selected_chromosomes]
			elif type(ga.fitness_type) == tuple or type(ga.fitness_type) == list:
				if ga.fitness_type[0] == 'equal':
					selected_chromosomes = mapped_chromosomes_rdd.takeOrdered(len(ga.population.members)-math.ceil(ga.cross_prob * len(ga.population.members)),key=lambda x: abs(x[1]-ga.fitness_type[1]))
					ga.best_fitness = selected_chromosomes[0]
					selected_chromosomes = [ x[0] for x in selected_chromosomes]
		else:
			selected_chromosomes = ga.handle_selection()

		ga.population.new_members = selected_chromosomes

		if ga.fitness_type[0] == 'equal':
			if ga.best_fitness[1] == ga.fitness_type[1]:
				return 1

		n = math.ceil(ga.cross_prob * len(ga.population.members))
		if n %2 == 1:
			n -= 1
			ga.population.new_members.append(ga.population.members[random.randint(0,len(ga.population.members)-1)])

		total_fitness = mapped_chromosomes_rdd.map(lambda x: (x[0],x[1]) if x[1] > 0 else (x[0],random.uniform(0.01, 0.02))).values().sum()

		p = mapped_chromosomes_rdd.map(lambda x: (x[0],x[1]/total_fitness)).values().collect()

		p[random.randint(0,len(p)-1)] += (1-sum(p))


		# Crossover Mapping
		crossover_indexes = np.random.choice(len(ga.population.members),n,p=p, replace=False)
		# print("crossover_indices = ",crossover_indexes)
		crossover_chromosomes = [ ga.population.members[index] for index in crossover_indexes]
		crossover_pair_indexes = [(crossover_indexes[i],crossover_indexes[i+1]) for i in range(0,len(crossover_indexes),2)]

		crossover_pair_indexes_rdd = sc.parallelize(crossover_pair_indexes)
		crossover_before = crossover_pair_indexes_rdd.map(lambda x: (ga.population.members[0],ga.population.members[1]))
		crossover_after = crossover_before.map(lambda x:(x,ga.chooseCrossoverHandler()(x[0],x[1])))

		crossover_results = crossover_after.flatMap(lambda x:x[1]).collect()
		ga.population.new_members.extend(crossover_results)
		
		mutation_indexes = np.random.choice(len(ga.population.new_members),int(ga.mut_prob*len(p)), replace=False)
		mutation_indexes_rdd = sc.parallelize(mutation_indexes)
		mutation_before = mutation_indexes_rdd.map(lambda x:(x,ga.population.new_members[x]))
		mutation_results = mutation_before.map(lambda x:(x[0],x[1],ga.chooseMutationHandler()(list(x[1])))).collect()

		for entry in mutation_results:
			ga.population.new_members[entry[0]] = entry[2]

	def evolve(self,ga):
		"""
		Invokes the private method __evolve_normal to perform an iteration Genetic Algorithm

		Parameters :
		-------------
		ga : reference to the GAEngine object

		Returns : 1 if optimal solution was found

		"""
		if self.pyspark == False:
			return self.__evolve_normal(ga)
		else:
			return self.__evolve_pyspark(ga)



class Population:
	"""
	Class contians info on population of candidate solutions

	Instance Members :
	----------------
	population_size : int
	members : List containing the members of population 
	new_members : List containing members of population after each iteration

	Methods :
	----------
	createMembers() : Generates initial members of population by invoking one of the methods from ChromosomeFactory.py module

	"""

	def __init__(self,factory,population_size):
		self.population_size = population_size
		self.members = []
		self.new_members = []
		self.createMembers(factory)

	def createMembers(self,factory):
		"""
		Generates initial members of population by invoking one of the methods from
		ChromosomeFactory.py module

		Parameters :
		-------------
		factory : Reference to a method from ChromosomeFactory.py module

		"""

		for i in range(self.population_size):
			self.members.append(factory.createChromosome())




from pygenetic import Population, Evolution, Statistics
import random
import collections
import bisect
import math
import numpy as np

class GAEngine:
	"""
	This Class is the main driver program which contains and invokes the operators used in Genetic algorithm

	GAEngine keeps track of specific type of operators the user has specified for running the algorithm

	Methods :
	---------
	addCrossoverHandler(crossover_handler, weight, *args)
		Sets the function to be used for crossover operation

	addMutationHandler(mutation_handler, weight, *args)
		Sets the function to be used for mutation operation

	doCrossover(cross_func, member1, member2)
		Calls crossover handler passing 2 given members as parameters

	doMutation(self, mut_func, member)
		Calls mutation handler passing member as parameter

	setSelectionHandler(selection_handler, *args)
		Sets the function to be used for selection operation

	setFitnessHandler(fit_function, *args)
		Sets the function to be used for calculating fitness of chromosome

	calculateFitness(chromosome)
		Invokes fitness function (fitness_func) to compute the fitness score of a chromosome

	generateFitnessMappings()
		Generates a list of tuples (individual, fitness_score) and also stores the tuple
		containing fittest chromosome [best_fitness] depending on fitness_type(max/min/equal)

	handle_selection(repeat_chromosome_sorting=False)
		Invokes generateFitnessMappings() if repeat_chromosome_sorting=True and  selection_handler called

	normalizeWeights()
		Normalizes crossover and mutation handler weights, result is a CDF

	chooseCrossoverHandler()
		Selects crossover handler from available handlers based on weightage given to handlers

	chooseMutationHandler()
		Selects mutation handler from available handlers based on weightage given to handlers

	setEvolution(evolution)
		Sets evolution instance member to the parameter passed during class initialization

	addStatistic(statistic,statistic_function)
		Appends a new statistic function

	evolve(noOfIterations=50)
		The interface provided to start evolution

	continue_evolve(noOfIterations=20)
		Performs the evolution by invoking the evolve method from Evolution.py module
		as many times as number of iterations specified by user or terminates if optimal
		solution is found.
		Also invokes compute method from Statistics.py module to generate graph


	Instance Members
	-----------
	fitness_func : A function reference
				The fitness function to be used, passed as a function argument

	factory : Instance of any subclass of ChromosomeFactory class
				Generates and returns the initial population of candidate solutions

	population_size : int
				The number of candidate solutions that can exist after every iteration

	cross_prob : float (0.0 to 1.0)
				The Crossover probability of crossover operation which determines the extent to which crossover between parents
				(default is 0.7)

	mut_prob : float (0.0 to 1.0)
				The mutation probability of mutation operation which determines extent to which candidates should be mutated
				(default is 0.1)

	fitness_type : string
				Indicates the nature of fitness value (higher/lower/equal) to be considered during selection of candidates
				(default is max)

	adaptive_mutation : boolean
				If set rate of mutation of candidates dynamically changes during execution depending on diversity in population
				(default is true)

	initial_mut_prob : float
				If adaptive mutation is True, this stores the initial mutation probability

	crossover_handlers : list
				Stores all crossover handlers added to solve problem

	crossover_handlers_weights : list
				Stores the weights associated with crossover handlers in crossover_handlers. Sums up to 1

	mutation_handlers : list
				Stores all mutation handlers added to solve problem

	mutation_handlers_weights : list
				Stores the weights associated with mutation handlers in mutation_handlers. Sums up to 1

	selection_handler : A function reference
				Stores the function selected by user to perform selection

	fitness_mappings : list
				List containing population and associated fitness values, refreshed each generation

	population : Class Instance
				An instance of the class "Population" used to create initial population.

	best_fitness : tuple
				Stores the best fitness for each generation

	evolution : Class Instance
				An instance of any class that supports implementation of evolution

	population_control : boolean
				If true, population size is maintained. Else, population size may vary in each iteration

	hall_of_fame_injection : boolean
				If true, best chromosome among all past generations is injected into population after a set number of generations

	efficient_iteration_halt : boolean
				If true, halts evolution when the best chromosome remains the same for a consecutive number of generations

	fitness_external_data : list
				Stores any additional data structures required for fitness calculation

	selection_external_data : list
				Stores any additional data structures required to perform selection

	crossover_external_data : dict
				Stores any additional data required to perform crossover. Data is mapped to specific crossover function

	mutation_external_data : dict
				Stores any additional data required to perform mutation. Data is mapped to specific mutation function

	hall_of_fame : tuple
				Stores the best chromosome among all the chromosomes generated in all past generations

	extra_statistics : dict
				Stores any extra statistics added by user

  	"""


	def __init__(self,factory,population_size=100,cross_prob=0.7,mut_prob=0.1,fitness_type='max',adaptive_mutation=True, population_control=False,hall_of_fame_injection=True,efficient_iteration_halt=False,use_pyspark=False):
		self.fitness_func = None
		self.factory = factory
		self.cross_prob = cross_prob
		self.adaptive_mutation = adaptive_mutation
		if self.adaptive_mutation:
			self.initial_mut_prob = mut_prob
		else:
			self.mut_prob = mut_prob
		self.crossover_handlers = []
		self.crossover_handlers_weights = []
		self.mutation_handlers = []
		self.mutation_handlers_weights = []
		self.selection_handler = None
		self.fitness_type = fitness_type
		self.fitness_mappings = None
		self.population = None
		self.population_size = population_size
		if type(self.fitness_type) == str:
			if self.fitness_type == 'max':
				self.best_fitness = None, float("-inf")
			elif self.fitness_type == 'min':
				self.best_fitness = None, float("inf")
			else:
				raise Exception('Invalid Fitness Type given to GAEngine')
		elif type(self.fitness_type) == tuple or type(self.fitness_type) == list:
			if self.fitness_type[0] == 'equal':
				self.best_fitness = None, float("inf")
		else:
			raise Exception('Invalid Fitness Type given to GAEngine')
		self.evolution = Evolution.StandardEvolution(pyspark=use_pyspark)
		self.population_control = population_control
		self.hall_of_fame_injection = hall_of_fame_injection
		self.efficient_iteration_halt = efficient_iteration_halt
		self.fitness_external_data = []
		self.selection_external_data = []
		self.crossover_external_data = {}
		self.mutation_external_data = {}
		self.hall_of_fame = None
		self.extra_statistics = {}

	def addCrossoverHandler(self,crossover_handler, weight = 1, *args):
		"""
		Adds crossover handler staticmethod defined in Utils.py and
		appends the weightage to be given to the handler

		Parameters :
		----------
		crossover_handler : Method defined in Utils.py or custom
		weight : integer relative weightage
		any number of extra data required for specific handler

		"""

		self.crossover_handlers.append(crossover_handler)
		self.crossover_handlers_weights.append(weight)
		xtra_args = []
		for arg in args:
			xtra_args.append(arg)
		self.crossover_external_data.update({crossover_handler:tuple(xtra_args)})

			

	def addMutationHandler(self,mutation_handler, weight = 1, *args):
		"""
		Adds mutation handler staticmethod defined in Utils.py and
		appends the weightage to be given to the handler

		Parameters :
		----------
		mutation_handler : Method defined in Utils.py or custom
		weight : integer relative weightage
		any number of extra data required for specific handler

		"""

		self.mutation_handlers.append(mutation_handler)
		self.mutation_handlers_weights.append(weight)
		xtra_args = []
		for arg in args:
			xtra_args.append(arg)
		self.mutation_external_data.update({mutation_handler:tuple(xtra_args)})

	def doCrossover(self, cross_func, member1, member2):
		"""
		Calls crossover handler passing 2 given members as parameters

		Parameters :
		----------
		cross_func : Crossover handler to be used
		member1 : Parent 1
		member2 : Parent 2

		Returns :
		----------
		Tuple containing two children chromosomes resulting from crossover between passed parents
		"""
		if cross_func in self.crossover_external_data:
			return cross_func(member1, member2, *(self.crossover_external_data[cross_func]))
		else:
			return cross_func(member1, member2)

	def doMutation(self, mut_func, member):
		"""
		Calls mutation handler passing a member as parameter

		Parameters :
		----------
		mut_func : mutation handler to be used
		member : The chromosome to perform mutation on

		Returns :
		----------
		A chromosome resulting from mutation on passed member
		"""
		if mut_func in self.mutation_external_data:
			return mut_func(member, *(self.mutation_external_data[mut_func]))
		else:
			return mut_func(member)

	def setSelectionHandler(self,selection_handler, *args):
		"""
		Sets function to be used for selection_handler

		Parameters:
		----------
		selection_handler : Function to be used to perform selection, can be custom
		any extra data required to do selection 

		"""
		self.selection_handler = selection_handler
		for arg in args:
			self.selection_external_data.append(arg)

	def setFitnessHandler(self, fit_function, *args):
		"""
		Sets function to be used to calculate fitness

		Parameters:
		----------
		fit_function : Function to be used to calculate fitness of a chromosome, can be custom
		any extra data required to calculate fitness 

		"""
		self.fitness_func = fit_function
		for arg in args:
			self.fitness_external_data.append(arg)

	def calculateFitness(self,chromosome):
		"""
		Invokes fitness function (fitness_func) to compute the fitness score of a chromosome

		Parameters:
		----------
		chromosome for which fitness is to be calculated

		Returns:
		--------
		Fitness value of chromosome

		"""
		if self.fitness_external_data:
			return self.fitness_func(chromosome, *(self.fitness_external_data))
		else:
			return self.fitness_func(chromosome)

	def generateFitnessMappings(self):
		"""
		Generates a list of tuples (individual, fitness_score) and also stores the tuple
		containing fittest chromosome [best_fitness] depending on fitness_type(max/min/equal)

		"""

		self.fitness_mappings = [(member, self.calculateFitness(member)) for member in self.population.members]
		if type(self.fitness_type) == str:
			if self.fitness_type == 'max':
				self.fitness_mappings.sort(key=lambda x:x[1],reverse=True)
				self.best_fitness = self.fitness_mappings[0]
				if self.hall_of_fame:
					if self.best_fitness[1] > self.hall_of_fame[1]:
						self.hall_of_fame = self.best_fitness
				else:
					self.hall_of_fame = self.best_fitness

			elif self.fitness_type == 'min':
				self.fitness_mappings.sort(key=lambda x:x[1])
				self.best_fitness = self.fitness_mappings[0]
				if self.hall_of_fame:
					if self.best_fitness[1] < self.hall_of_fame[1]:
						self.hall_of_fame = self.best_fitness
				else:
					self.hall_of_fame = self.best_fitness

		elif type(self.fitness_type) == tuple or type(self.fitness_type) == list:
			self.fitness_mappings.sort(key=lambda x:abs(x[1]-self.fitness_type[1]))
			self.best_fitness = self.fitness_mappings[0]
			if self.hall_of_fame:
				if abs(self.fitness_type[1] - self.best_fitness[1]) < abs(self.fitness_type[1] - self.hall_of_fame[1]):
					self.hall_of_fame = self.best_fitness
			else:
				self.hall_of_fame = self.best_fitness

	def handle_selection(self,repeat_chromosome_sorting=False):

		"""
		Invokes generateFitnessMappings() if repeat_chromosome_sorting=True to generate list of (chromosome,fitness)
		Invokes selection_handler staticmethod defined in Utils.py module or custom

		Parameters :
		----------
		repeat_chromosome_sorting : To sort before selection or not

		Returns :
		---------
		List of limited number of fittest members of population

		"""
		if repeat_chromosome_sorting:
			self.generateFitnessMappings()
		if self.selection_external_data:
			return self.selection_handler(self.fitness_mappings,self, *(self.selection_external_data))
		else:
			return self.selection_handler(self.fitness_mappings,self)

	def normalizeWeights(self):
		"""
		Normalizes the weights of mutation and crossover handlers

		"""

		# Normalizing crossover and mutation handler weights, result is a CDF
		total = sum(self.mutation_handlers_weights)
		cumsum = 0
		for i in range(len(self.mutation_handlers_weights)):
			cumsum += self.mutation_handlers_weights[i]
			self.mutation_handlers_weights[i] = cumsum/total
		total = sum(self.crossover_handlers_weights)
		cumsum = 0
		for i in range(len(self.crossover_handlers_weights)):
			cumsum += self.crossover_handlers_weights[i]
			self.crossover_handlers_weights[i] = cumsum/total

	def chooseCrossoverHandler(self):
		"""
		Selects crossover handler from available handlers based on weightage given to handlers

		Returns :
		--------
		The selected crossover handler function

		"""

		x = random.random()
		idx = bisect.bisect(self.crossover_handlers_weights, x)
		return self.crossover_handlers[idx]

	def chooseMutationHandler(self):
		"""
		Selects mutation handler from available handlers based on weightage given to handlers

		Returns :
		--------
		The selected mutation handler function

		"""

		x = random.random()
		idx = bisect.bisect(self.mutation_handlers_weights, x)
		return self.mutation_handlers[idx]

	def setEvolution(self,evolution):
		"""
		Sets evolution instance member to the parameter passed during class initialization

		Parameters :
		----------
		evolution : A predefined evoluton class instance or instance of custom class

		"""
		self.evolution = evolution

	def addStatistic(self,statistic,statistic_function):
		"""
		Appends a new statistic function

		Parameters :
		----------
		statistic : The statistic for which to add the function being passed
		statistic_function : The function to be added to the statistic

		"""
		if type(statistic) != str:
			raise Exception('Invalid Statistics key')
		self.extra_statistics[statistic] = statistic_function

	def evolve(self,noOfIterations=50):
		"""
		The interface provided to start evolution

		Parameters :
		-----------
		noOfIterations : int
						default value : 50

		"""
		self.population = Population.Population(self.factory,self.population_size)
		self.statistics = Statistics.Statistics()
		self.last_20_fitnesses = collections.deque([])
		self.continue_evolve(noOfIterations)
		
	def continue_evolve(self, noOfIterations=20):
		"""
		Performs the evolution by invoking the evolve method from Evolution.py module
		as many times as number of iterations specified by user or terminates if optimal
		solution is found.
		Also invokes compute method from Statistics.py module to generate graph

		Parameters :
		-----------
		noOfIterations : int
						default value : 20

		"""
		self.normalizeWeights()
		if self.population == None:
			raise Exception('Call evolve before calling continue_evolve')
		print("gen\tavg\t\tbest\tworst\t")
		i=0
		while i<noOfIterations:
			self.generateFitnessMappings()
			fitnesses = [ x[1] for x in self.fitness_mappings]
			self.statistics.add_statistic('best-fitness',self.fitness_mappings[0][1])
			self.statistics.add_statistic('worst-fitness',self.fitness_mappings[-1][1])
			self.mean_fitness = sum(fitnesses)/len(fitnesses)
			self.statistics.add_statistic('avg-fitness',self.mean_fitness)
			self.diversity = math.sqrt(sum((fitness - self.mean_fitness)**2 for fitness in fitnesses)) / len(fitnesses)
			print("%i\t%.2f\t\t%s\t%s\t" % (len(self.statistics.statistic_dict['best-fitness']),self.mean_fitness,self.fitness_mappings[0][1],self.fitness_mappings[-1][1]))
			if self.adaptive_mutation:
				self.mut_prob = self.initial_mut_prob * ( 1 + ((self.best_fitness[1]-self.diversity) / (self.diversity+self.best_fitness[1]) ) )
				self.mut_prob = np.clip(self.mut_prob,0.0001,0.8)
			self.statistics.add_statistic('mutation_rate',self.mut_prob)
			self.statistics.add_statistic('diversity',self.diversity)
			for statistic in self.extra_statistics:
				self.statistics.add_statistic(statistic,self.extra_statistics[statistic](self.fitness_mappings,self))

			result = self.evolution.evolve(self)

			if self.hall_of_fame_injection and (i+1)%20 == 0:
				self.population.new_members.insert(0,self.hall_of_fame[0]) 

			if self.population_control:
				if len(self.population.new_members) > self.population_size:
					self.population.new_members = self.population.new_members[:self.population_size]
				elif len(self.population.new_members) < self.population_size:
					self.population.new_members = self.population.new_members * int(self.population_size/len(self.population.new_members)) + self.population.new_members[:self.population_size%len(self.population.new_members)]

			if self.efficient_iteration_halt:
				if len(self.last_20_fitnesses)==20:
					self.last_20_fitnesses.popleft()
					self.last_20_fitnesses.append(self.best_fitness[1])
					if all(x == self.last_20_fitnesses[0] for x in self.last_20_fitnesses):
						break
				else:
					self.last_20_fitnesses.append(self.best_fitness[1])

			# For next iteration
			self.population.members = self.population.new_members
			self.population.new_members = []

			if result == 1:
				print('GA Problem Solved')
				break
			i += 1




import setuptools

with open("README.md", "r") as f:
   long_description = f.read()

setuptools.setup(
    name="pygenetic",
    version="1.0.2",
    author="Bharatraj S Telkar, Daniel Isaac, Shreyas V Patil",
    author_email="telkarraj@gmail.com, danielbcbs2@gmail.com, pshreyasv100@gmail.com",
    description="An Efficient Python Genetic Algorithm API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/danny311296/pygenetic",
    packages=['pygenetic'],
    include_package_data=True,
    license='MIT',
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    install_requires=[
          'rstr==2.2.6',
          'numpy==1.15.4',
          'matplotlib==2.2.2',
      ],
)